import '@nativescript/canvas-three';
import { DOCUMENT } from '@angular/common';
import * as i0 from '@angular/core';
import { input, booleanAttribute, output, inject, Injector, EnvironmentInjector, DestroyRef, NgZone, viewChild, ViewContainerRef, untracked, createEnvironmentInjector, Component, NO_ERRORS_SCHEMA, ChangeDetectionStrategy } from '@angular/core';
import { registerElement } from '@nativescript/angular';
import { Canvas } from '@nativescript/canvas';
import { NgtRoutedScene, injectStore, injectCanvasRootInitializer, makeDpr, provideNgtRenderer, provideStore } from 'angular-three';
import { WebGLRenderer } from 'three';

registerElement('Canvas', () => Canvas);
class NgtCanvasNative {
    constructor() {
        this.sceneGraph = input.required({
            transform: (value) => {
                if (value === 'routed')
                    return NgtRoutedScene;
                return value;
            },
        });
        this.gl = input();
        this.size = input();
        this.shadows = input(false, {
            transform: (value) => {
                if (value === '')
                    return booleanAttribute(value);
                return value;
            },
        });
        this.legacy = input(false, { transform: booleanAttribute });
        this.linear = input(false, { transform: booleanAttribute });
        this.flat = input(false, { transform: booleanAttribute });
        this.orthographic = input(false, { transform: booleanAttribute });
        this.frameloop = input('always');
        this.performance = input();
        this.dpr = input([1, 2]);
        this.raycaster = input();
        this.scene = input();
        this.camera = input();
        this.lookAt = input();
        this.created = output();
        this.store = injectStore();
        this.initRoot = injectCanvasRootInitializer();
        this.injector = inject(Injector);
        this.environmentInjector = inject(EnvironmentInjector);
        this.destroyRef = inject(DestroyRef);
        this.zone = inject(NgZone);
        this.canvasViewContainerRef = viewChild.required('canvas', { read: ViewContainerRef });
        this.destroyRef.onDestroy(() => {
            this.glRef?.destroy();
            this.glEnvironmentInjector?.destroy();
            this.configurator?.destroy();
        });
    }
    onReady(event) {
        const canvas = event.object;
        const dpr = makeDpr(untracked(this.dpr), window);
        const canvasWidth = canvas.clientWidth * dpr;
        const canvasHeight = canvas.clientHeight * dpr;
        Object.assign(canvas, { width: canvasWidth, height: canvasHeight });
        const context = canvas.getContext('webgl2');
        const gl = new WebGLRenderer({
            canvas,
            context: context,
            powerPreference: 'high-performance',
            antialias: true,
            alpha: true,
            ...untracked(this.gl),
        });
        gl.setSize(canvasWidth, canvasHeight);
        this.zone.runOutsideAngular(() => {
            this.configurator = this.initRoot(canvas);
            this.configurator.configure({
                gl,
                size: { width: canvasWidth, height: canvasHeight, top: 0, left: 0 },
                shadows: untracked(this.shadows),
                legacy: untracked(this.legacy),
                linear: untracked(this.linear),
                flat: untracked(this.flat),
                orthographic: untracked(this.orthographic),
                frameloop: untracked(this.frameloop),
                performance: untracked(this.performance),
                dpr: untracked(this.dpr),
                raycaster: untracked(this.raycaster),
                scene: untracked(this.scene),
                camera: untracked(this.camera),
                lookAt: untracked(this.lookAt),
            });
            untracked(this.noZoneRender.bind(this));
        });
    }
    noZoneRender() {
        // NOTE: destroy previous instances if existed
        this.glEnvironmentInjector?.destroy();
        this.glRef?.destroy();
        // NOTE: Flag the canvas active, rendering will now begin
        this.store.update((state) => ({ internal: { ...state.internal, active: true } }));
        // emit created event if observed
        this.created.emit(this.store.snapshot);
        this.glEnvironmentInjector = createEnvironmentInjector([{ provide: DOCUMENT, useValue: document }, provideNgtRenderer(this.store)], this.environmentInjector);
        this.glRef = untracked(this.canvasViewContainerRef).createComponent(untracked(this.sceneGraph), {
            environmentInjector: this.glEnvironmentInjector,
            injector: this.injector,
        });
        this.glRef.changeDetectorRef.detectChanges();
    }
    static { this.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "19.1.4", ngImport: i0, type: NgtCanvasNative, deps: [], target: i0.ɵɵFactoryTarget.Component }); }
    static { this.ɵcmp = i0.ɵɵngDeclareComponent({ minVersion: "17.2.0", version: "19.1.4", type: NgtCanvasNative, isStandalone: true, selector: "NgtCanvas", inputs: { sceneGraph: { classPropertyName: "sceneGraph", publicName: "sceneGraph", isSignal: true, isRequired: true, transformFunction: null }, gl: { classPropertyName: "gl", publicName: "gl", isSignal: true, isRequired: false, transformFunction: null }, size: { classPropertyName: "size", publicName: "size", isSignal: true, isRequired: false, transformFunction: null }, shadows: { classPropertyName: "shadows", publicName: "shadows", isSignal: true, isRequired: false, transformFunction: null }, legacy: { classPropertyName: "legacy", publicName: "legacy", isSignal: true, isRequired: false, transformFunction: null }, linear: { classPropertyName: "linear", publicName: "linear", isSignal: true, isRequired: false, transformFunction: null }, flat: { classPropertyName: "flat", publicName: "flat", isSignal: true, isRequired: false, transformFunction: null }, orthographic: { classPropertyName: "orthographic", publicName: "orthographic", isSignal: true, isRequired: false, transformFunction: null }, frameloop: { classPropertyName: "frameloop", publicName: "frameloop", isSignal: true, isRequired: false, transformFunction: null }, performance: { classPropertyName: "performance", publicName: "performance", isSignal: true, isRequired: false, transformFunction: null }, dpr: { classPropertyName: "dpr", publicName: "dpr", isSignal: true, isRequired: false, transformFunction: null }, raycaster: { classPropertyName: "raycaster", publicName: "raycaster", isSignal: true, isRequired: false, transformFunction: null }, scene: { classPropertyName: "scene", publicName: "scene", isSignal: true, isRequired: false, transformFunction: null }, camera: { classPropertyName: "camera", publicName: "camera", isSignal: true, isRequired: false, transformFunction: null }, lookAt: { classPropertyName: "lookAt", publicName: "lookAt", isSignal: true, isRequired: false, transformFunction: null } }, outputs: { created: "created" }, providers: [{ provide: DOCUMENT, useValue: document }, provideStore()], viewQueries: [{ propertyName: "canvasViewContainerRef", first: true, predicate: ["canvas"], descendants: true, read: ViewContainerRef, isSignal: true }], ngImport: i0, template: `
		<GridLayout>
			<Canvas #canvas style="width: 100%; height: auto" (ready)="onReady($event)"></Canvas>
		</GridLayout>
	`, isInline: true, changeDetection: i0.ChangeDetectionStrategy.OnPush }); }
}
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "19.1.4", ngImport: i0, type: NgtCanvasNative, decorators: [{
            type: Component,
            args: [{
                    selector: 'NgtCanvas',
                    standalone: true,
                    template: `
		<GridLayout>
			<Canvas #canvas style="width: 100%; height: auto" (ready)="onReady($event)"></Canvas>
		</GridLayout>
	`,
                    providers: [{ provide: DOCUMENT, useValue: document }, provideStore()],
                    schemas: [NO_ERRORS_SCHEMA],
                    changeDetection: ChangeDetectionStrategy.OnPush,
                }]
        }], ctorParameters: () => [] });

/**
 * Generated bundle index. Do not edit.
 */

export { NgtCanvasNative };
//# sourceMappingURL=angular-three-nativescript.mjs.map
