import { NgtCanvasElement, NgtGlobalRenderCallback, NgtState } from './types';
import { NgtSignalStore } from './utils/signal-store';
export declare const roots: Map<NgtCanvasElement, NgtSignalStore<NgtState>>;
/**
 * Adds a global render callback which is called each frame.
 * @see https://docs.pmnd.rs/react-three-fiber/api/additional-exports#addEffect
 */
export declare const addEffect: (callback: NgtGlobalRenderCallback) => () => void;
/**
 * Adds a global after-render callback which is called each frame.
 * @see https://docs.pmnd.rs/react-three-fiber/api/additional-exports#addAfterEffect
 */
export declare const addAfterEffect: (callback: NgtGlobalRenderCallback) => () => void;
/**
 * Adds a global callback which is called when rendering stops.
 * @see https://docs.pmnd.rs/react-three-fiber/api/additional-exports#addTail
 */
export declare const addTail: (callback: NgtGlobalRenderCallback) => () => void;
export type NgtGlobalEffectType = 'before' | 'after' | 'tail';
export declare function flushGlobalEffects(type: NgtGlobalEffectType, timestamp: number): void;
export declare const injectLoop: {
    (): {
        loop: (timestamp: number) => void;
        invalidate: (store?: NgtSignalStore<NgtState>, frames?: number) => void;
        advance: (timestamp: number, runGlobalEffects?: boolean, store?: NgtSignalStore<NgtState>, frame?: XRFrame) => void;
    };
    (injectOptions: import("@angular/core").InjectOptions & {
        optional?: false;
    } & {
        injector?: import("@angular/core").Injector;
    }): {
        loop: (timestamp: number) => void;
        invalidate: (store?: NgtSignalStore<NgtState>, frames?: number) => void;
        advance: (timestamp: number, runGlobalEffects?: boolean, store?: NgtSignalStore<NgtState>, frame?: XRFrame) => void;
    };
    (injectOptions: import("@angular/core").InjectOptions & {
        injector?: import("@angular/core").Injector;
    }): {
        loop: (timestamp: number) => void;
        invalidate: (store?: NgtSignalStore<NgtState>, frames?: number) => void;
        advance: (timestamp: number, runGlobalEffects?: boolean, store?: NgtSignalStore<NgtState>, frame?: XRFrame) => void;
    } | null;
};
