import { Signal } from '@angular/core';
import { Vector2, Vector3, Vector4 } from 'three';
import { NgtVector2, NgtVector3, NgtVector4 } from '../three-types';
type KeysOfType<TObject extends object, TType> = Exclude<{
    [K in keyof TObject]: TObject[K] extends TType | undefined | null ? K : never;
}[keyof TObject], undefined>;
export declare function omit<TObject extends object, TKeys extends (keyof TObject)[]>(objFn: () => TObject, keysToOmit: TKeys): Signal<Omit<TObject, TKeys[number]>>;
export declare function pick<TObject extends object, TKey extends keyof TObject>(objFn: () => TObject, key: TKey): Signal<TObject[TKey]>;
export declare function pick<TObject extends object, TKeys extends (keyof TObject)[]>(objFn: () => TObject, keys: TKeys): Signal<Pick<TObject, TKeys[number]>>;
export declare function merge<TObject extends object>(objFn: () => TObject, toMerge: Partial<TObject>, mode?: 'override' | 'backfill'): Signal<TObject & Partial<TObject>>;
type NgtVectorComputed<TVector extends Vector2 | Vector3 | Vector4, TNgtVector = TVector extends Vector2 ? NgtVector2 : TVector extends Vector3 ? NgtVector3 : NgtVector4> = {
    (input: Signal<TNgtVector>): Signal<TVector>;
    (input: Signal<TNgtVector>, keepUndefined: true): Signal<TVector | undefined>;
    <TObject extends object>(options: Signal<TObject>, key: KeysOfType<TObject, TNgtVector>): Signal<TVector>;
    <TObject extends object>(options: Signal<TObject>, key: KeysOfType<TObject, TNgtVector>, keepUndefined: true): Signal<TVector | undefined>;
};
export declare const vector2: NgtVectorComputed<Vector2, NgtVector2>;
export declare const vector3: NgtVectorComputed<Vector3, NgtVector3>;
export declare const vector4: NgtVectorComputed<Vector4, NgtVector4>;
export {};
