"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.addSobaGenerator = addSobaGenerator;
const devkit_1 = require("@nx/devkit");
const enquirer_1 = require("enquirer");
const utils_1 = require("../utils");
const version_1 = require("../version");
const ENTRY_POINTS = {
    abstractions: ['troika-three-text'],
    controls: ['camera-controls', 'maath'],
    materials: ['three-custom-shader-material'],
    staging: ['troika-three-text', '@monogrid/gainmap-js'],
    stats: ['stats-gl'],
};
async function addSobaGenerator(tree) {
    const packagesToAdd = ['three-stdlib', '@pmndrs/vanilla', 'three-mesh-bvh'];
    const packageJson = (0, devkit_1.readJson)(tree, 'package.json');
    const ngtVersion = packageJson['dependencies']['angular-three'] || packageJson['devDependencies']['angular-three'];
    if (!ngtVersion) {
        devkit_1.logger.error('angular-three is not installed.');
        return;
    }
    devkit_1.logger.info('Adding angular-three-soba...');
    (0, utils_1.addMetadataJson)(tree, 'angular-three-soba/metadata.json');
    const { peerDependencies } = await (0, enquirer_1.prompt)({
        type: 'multiselect',
        name: 'peerDependencies',
        message: `To know which peer dependencies we need to add, please select the secondary entry points you are planning to use:`,
        choices: [
            ...Object.entries(ENTRY_POINTS).map(([entryPoint, deps]) => ({
                value: deps,
                name: `angular-three-soba/${entryPoint}`,
                message: entryPoint,
            })),
            {
                value: [
                    'troika-three-text',
                    'three-custom-shader-material',
                    '@monogrid/gainmap-js',
                    'camera-controls',
                    'maath',
                    'stats-gl',
                ],
                name: 'All',
                message: 'I am not sure. Let me add all the peer dependencies.',
            },
        ],
        // @ts-expect-error - result is typed for single select but we're using multi select
        result(values) {
            const mapped = this.map(values);
            return values.flatMap((value) => mapped[value]);
        },
    });
    // flatten, dedupe peerDependencies, add to packagesToAdd
    peerDependencies
        .filter((item, index, array) => array.indexOf(item) === index)
        .forEach((item) => {
        if (!packagesToAdd.includes(item)) {
            packagesToAdd.push(item);
        }
    });
    const depsToAdd = packagesToAdd.reduce((acc, item) => {
        if (version_1.SOBA_PEER_DEPENDENCIES[item]) {
            acc[item] = version_1.SOBA_PEER_DEPENDENCIES[item];
        }
        return acc;
    }, { 'angular-three-soba': ngtVersion });
    // add deps to package.json
    devkit_1.logger.info('Adding dependencies to package.json...');
    (0, devkit_1.addDependenciesToPackageJson)(tree, depsToAdd, {});
    await (0, devkit_1.formatFiles)(tree);
    return () => {
        (0, devkit_1.installPackagesTask)(tree);
    };
}
exports.default = addSobaGenerator;
//# sourceMappingURL=generator.js.map