"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.addMetadataJson = addMetadataJson;
const devkit_1 = require("@nx/devkit");
function addMetadataJson(tree, metadataJsonPath) {
    if (!metadataJsonPath.includes('node_modules')) {
        metadataJsonPath = `./node_modules/${metadataJsonPath}`;
    }
    // add metadata.json to vscode settings if exists
    const vscodeSettingsPath = '.vscode/settings.json';
    if (tree.exists('.vscode')) {
        devkit_1.logger.info('Enabling typings support for VSCode...');
        if (!tree.exists(vscodeSettingsPath)) {
            (0, devkit_1.writeJson)(tree, vscodeSettingsPath, {});
        }
        (0, devkit_1.updateJson)(tree, vscodeSettingsPath, (json) => {
            if (json['html.customData'] &&
                Array.isArray(json['html.customData']) &&
                !json['html.customData'].includes(metadataJsonPath)) {
                json['html.customData'].push(metadataJsonPath);
            }
            else {
                json['html.customData'] = [metadataJsonPath];
            }
            return json;
        });
    }
    else {
        devkit_1.logger.info(`.vscode/settings.json not found.
If you are using VSCode, add "${metadataJsonPath}" to "html.customData" in ".vscode/settings.json"
to enable TypeScript type definitions for Angular Three elements.
`);
    }
}
//# sourceMappingURL=utils.js.map